#include <stdio.h>
#include <stdlib.h>

#include "bst.h"

// Private functions.
void bst_node_walk_inorder(struct bst_node * node, void (*op)(void * data));
void * bst_node_search(struct bst * bst, struct bst_node * node, void * data);
void * bst_node_minimum(struct bst_node * node);
void * bst_node_maximum(struct bst_node * node);

////////////////////////////////////////////////////////////////////////////////
struct bst *
bst_new(bool (*le)(void * a, void * b),
	bool (*eq)(void * a, void * b))
   {
   if ( (NULL == le) || (NULL == eq) )
      return NULL;

   struct bst * bst = malloc(sizeof(struct bst));
   if (NULL == bst)
      return NULL;

   bst->root = NULL;
   bst->size = 0;
   bst->le = le;
   bst->eq = eq;
   return bst;
   }


////////////////////////////////////////////////////////////////////////////////
void
bst_delete(struct bst * bst)
   {
   if (NULL == bst) return;
   if (0 < bst->size)
      {
      // Walk the tree and delete the nodes.
      // This will cause memory leaks, so hopefully the user has already
      // emptied the tree.
      }
   free(bst);
   }

////////////////////////////////////////////////////////////////////////////////
void *
bst_insert(struct bst * bst, void * data)
   {
   if ( (NULL == bst) || (NULL == data) )
      return NULL;

   struct bst_node * new = malloc(sizeof(struct bst_node));
   if (NULL == new)
      return NULL;
   new->left = new->right = new->parent = NULL;
   new->data = data;

   if (NULL == bst->root)
      {
      bst->root = new;
      return new->data;
      }

   struct bst_node * p = NULL;
   struct bst_node * c = bst->root;

   while (NULL != c)
      {
      p = c;
      if (bst->le(new->data, c->data))
	 c = c->left;
      else
	 c = c->right;
      }
   new->parent = p;
   if (bst->le(new->data, p->data))
      p->left = new;
   else
      p->right = new;

   return new->data;
   }


////////////////////////////////////////////////////////////////////////////////
void bst_walk_inorder(struct bst * bst, void (*op)(void * data))
   {
   if ( (NULL == bst) || (NULL == bst->root) || (NULL == op) )
      return;

   bst_node_walk_inorder(bst->root, op);
   }

////////////////////////////////////////////////////////////////////////////////
void bst_node_walk_inorder(struct bst_node * node, void (*op)(void * data))
   {
   if (NULL != node)
      {
      bst_node_walk_inorder(node->left, op);
      op(node->data);
      bst_node_walk_inorder(node->right, op);
      }
   }

////////////////////////////////////////////////////////////////////////////////
void * bst_search(struct bst * bst, void * data)
   {
   if ( (NULL == bst) || (NULL == bst->root) || (NULL == data) )
      return NULL;

   return bst_node_search(bst, bst->root, data);
   }

////////////////////////////////////////////////////////////////////////////////
// Recursive
void * bst_node_search(struct bst * bst, struct bst_node * node, void * data)
   {
   if (NULL == node)
      return NULL;

   if (bst->eq(data, node->data))
      return node->data;

   if (bst->le(data, node->data))
      return bst_node_search(bst, node->left, data);
   else
      return bst_node_search(bst, node->right, data);
   }

////////////////////////////////////////////////////////////////////////////////
// Iterative
void * bst_search_iterative(struct bst * bst, void * data)
   {
   if ( (NULL == bst) || (NULL == bst->root) || (NULL == data) )
      return NULL;

   struct bst_node * node = bst->root;
   while (1)
      {
      if (NULL == node)
	 return NULL;
      
      if (bst->eq(data, node->data))
	 return node->data;
      
      if (bst->le(data, node->data))
	 node = node->left;
      else
	 node = node->right;
      }
   }

////////////////////////////////////////////////////////////////////////////////
void * bst_minimum(struct bst * bst)
   {
   if (NULL == bst)
      return NULL;

   return bst_node_minimum(bst->root);
   }

////////////////////////////////////////////////////////////////////////////////
void * bst_node_minimum(struct bst_node * node)
   {
   while ( (NULL != node) && (NULL != node->left) )
      node = node->left;
   if (NULL == node)
      return NULL;
   return node->data;
   }

////////////////////////////////////////////////////////////////////////////////
void * bst_maximum(struct bst * bst)
   {
   if (NULL == bst)
      return NULL;

   return bst_node_maximum(bst->root);
   }

////////////////////////////////////////////////////////////////////////////////
void * bst_node_maximum(struct bst_node * node)
   {
   while ( (NULL != node) && (NULL != node->right) )
      node = node->right;
   if (NULL == node)
      return NULL;
   return node->data;
   }
